import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

enum SNSMobileSDKStatus {
  Ready,
  Initial,
  Incomplete,
  Pending,
  Approved,
  Failed,
  FinallyRejected,
  TemporarilyDeclined
}

enum SNSMobileSDKErrorType {
  Unknown,
  InvalidParameters,
  Unauthorized,
  InitialLoadingFailed,
  ApplicantNotFound,
  ApplicantMisconfigured,
  NetworkError,
  UnexpectedError,
}

typedef SNSTokenExpirationHandler = Future<String> Function();
typedef SNSReady = Function();
typedef SNSStatusChangedHandler = Function(SNSMobileSDKStatus newStatus, SNSMobileSDKStatus oldStatus);
typedef SNSErrorHandler = Function(SNSException exception);

class SNSMobileSDK {

  MethodChannel _channel = MethodChannel('sumsub.com/flutter_idensic_mobile_sdk_plugin');

  static SNSMobileSDKBuilder builder(String apiUrl, String flowName) => SNSMobileSDKBuilder(apiUrl: apiUrl, flowName: flowName);

  final String apiUrl;
  final String flowName;
  final String accessToken;
  final Locale locale;
  final String supportEmail;
  final SNSTokenExpirationHandler onTokenExpiration;
  final SNSStatusChangedHandler onStatusChanged;
  final bool isDebug;

  SNSMobileSDK._builder(SNSMobileSDKBuilder builder) :
    apiUrl = builder.apiUrl,
    flowName = builder.flowName,
    accessToken = builder.accessToken,
    locale = builder.locale,
    supportEmail = builder.supportEmail,
    onTokenExpiration = builder.onTokenExpiration,
    onStatusChanged = builder.onStatusChanged,
    isDebug = builder.isDebug;

  void dismiss() {
    _channel.invokeMethod('dismiss');
  }

  Future<SNSMobileSDKResult> launch() async {
    _channel.setMethodCallHandler(_onMethodCallHandler);
    final Map<dynamic, dynamic> map = await _channel.invokeMethod('onLaunchSDK', _toArguments());
    return Future.value(SNSMobileSDKResult.fromMap(map));
  }

  Future<String> _onMethodCallHandler(MethodCall call) async {
    switch(call.method) {
      case "onTokenExpiration":
        return onTokenExpiration();
      case "onError":
        return null;
      case "onStatusChanged":
        return onStatusChanged(toStatus(call.arguments[0]), toStatus(call.arguments[1]));
      default:
        print('Unknown method ${call.method}');
        throw MissingPluginException();
    }
  }

  Map<String, dynamic> _toArguments() {
    return {
      "apiUrl": apiUrl,
      "flowName": flowName,
      "accessToken": accessToken,
      "languageCode": locale?.languageCode,
      "supportEmail": supportEmail,
      "isDebug": isDebug,
    };
  }
}

class SNSMobileSDKBuilder {
  final String apiUrl;
  final String flowName;

  // builder
  String accessToken;
  Locale locale;
  String supportEmail;
  SNSTokenExpirationHandler onTokenExpiration;
  SNSStatusChangedHandler onStatusChanged;
  bool isDebug;

  SNSMobileSDKBuilder({@required this.apiUrl, @required this.flowName});

  SNSMobileSDKBuilder withAccessToken(String accessToken, SNSTokenExpirationHandler onTokenExpiration) {
    this.accessToken = accessToken;
    this.onTokenExpiration = onTokenExpiration;
    return this;
  }

  SNSMobileSDKBuilder withLocale(Locale locale) {
    this.locale = locale;
    return this;
  }

  SNSMobileSDKBuilder withSupportEmail(String supportEmail) {
    this.supportEmail = supportEmail;
    return this;
  }

  SNSMobileSDKBuilder withHandlers({SNSStatusChangedHandler onStatusChanged}) {
    this.onStatusChanged = onStatusChanged;
    return this;
  }

  SNSMobileSDKBuilder withDebug(bool isDebug) {
    this.isDebug = isDebug;
    return this;
  }

  SNSMobileSDK build() => SNSMobileSDK._builder(this);
}

class SNSMobileSDKResult {

  final bool success;
  final SNSMobileSDKStatus status;
  final SNSMobileSDKErrorType errorType;
  final String errorMsg;

  SNSMobileSDKResult(this.success, this.status, this.errorType, this.errorMsg);

  static SNSMobileSDKResult fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSMobileSDKResult(map['success'] as bool, toStatus(map['status']), toErrorType(map['errorType']), map['errorMsg']);
  }

  String toString() => 'SNSMobileSDKResult(success: $success, status: $status, errorType: $errorType, errorMsg: $errorMsg)';
}

class SNSException {

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    switch (map['type']) {
      case "Api":
        return SNSApiException.fromMap(map['payload']);
      case "Network":
        return SNSNetworkException.fromMap(map['payload']);
      case "Unknown":
        return SNSUnknownException.fromMap(map['payload']);
    }

    throw new Exception("Unknown exception type (${map['type']}");
  }
}

class SNSApiException extends SNSException {

  final int code;
  final String description;
  final String correlationId;

  SNSApiException(this.code, this.description, this.correlationId);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSApiException(map['code'] as int, map['description'] as String, map['correlationId'] as String);
  }  

  @override
  String toString() => 'SNSApiException(code: $code, description: $description, correlationId: $correlationId)';
}

class SNSNetworkException extends SNSException {

  final String message;

  SNSNetworkException(this.message);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSNetworkException(map['message'] as String);
  }

  @override
  String toString() => 'SNSNetworkException(message: $message)';
}

class SNSUnknownException extends SNSException {

  final String message;

  SNSUnknownException(this.message);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSUnknownException(map['message'] as String);
  }

  @override
  String toString() => 'SNSUnknownException(message: $message)';
}

SNSMobileSDKStatus toStatus(String value) {
  SNSMobileSDKStatus status;
  switch (value) {
    case 'Ready': 
      status = SNSMobileSDKStatus.Ready;
      break;
    case 'Initial': 
      status = SNSMobileSDKStatus.Initial;
      break;
    case 'Incomplete': 
      status = SNSMobileSDKStatus.Incomplete;
      break;
    case 'Pending': 
      status = SNSMobileSDKStatus.Pending;
      break;
    case 'Approved': 
      status = SNSMobileSDKStatus.Approved;
      break;
    case 'Failed': 
      status = SNSMobileSDKStatus.Failed;
      break;
    case 'FinallyRejected': 
      status = SNSMobileSDKStatus.FinallyRejected;
      break;
    case 'TemporarilyDeclined': 
      status = SNSMobileSDKStatus.TemporarilyDeclined;
      break;
    default:
      break;
    }

    return status;
}

SNSMobileSDKErrorType toErrorType(String value) {
  SNSMobileSDKErrorType errorType;
  switch (value) {
    case 'Unknown': 
      errorType = SNSMobileSDKErrorType.Unknown;
      break;
    case 'InvalidParamaters': 
      errorType = SNSMobileSDKErrorType.InvalidParameters;
      break;
    case 'Unauthorized': 
      errorType = SNSMobileSDKErrorType.Unauthorized;
      break;
    case 'InitialLoadingFailed': 
      errorType = SNSMobileSDKErrorType.InitialLoadingFailed;
      break;
    case 'ApplicantNotFound': 
      errorType = SNSMobileSDKErrorType.ApplicantNotFound;
      break;
    case 'ApplicantMisconfigured': 
      errorType = SNSMobileSDKErrorType.ApplicantMisconfigured;
      break;
    case 'NetworkError': 
      errorType = SNSMobileSDKErrorType.NetworkError;
      break;
    case 'UnexpectedError': 
      errorType = SNSMobileSDKErrorType.UnexpectedError;
      break;
    default:
      break;
  }

  return errorType;
}